program FLC2BRUN;

{$A+,B+,D+,E-,F-,G+,I+,L+,N-,O-,P-,Q+,R+,S+,T-,V+,X+}
{$M 16384,131072,131072}

uses
  Objects;


type
  FLCFileHeader = record
    FileSize       : longint;
    Signature      : word;
    Frames         : word;
    Width          : word;
    Height         : word;
    Depth          : word;
    Status         : word;
    Delay          : word;
    Unused1        : longint;
    CTime          : array [1..4] of byte;
    CPrgInfo       : array [1..4] of byte;
    MTime          : array [1..4] of byte;
    MPrgInfo       : array [1..4] of byte;
    XScale         : word;
    YScale         : word;
    Unused2        : array [1..38] of byte;
    FirstFramePos  : longint;
    SecondFramePos : longint;
    Unused3        : array [1..40] of byte;
  end;

  FLCFrameHeader = record
    Size      : longint;
    Signature : word;
    Chunks    : word;
    Unused    : array [1..8] of byte;
  end;

  FLCChunkHeader = record
    Size      : longint;
    Signature : word;
  end;


const
  flc_Palette256      = $04;
  flc_SS2             = $07;
  fli_Palette64       = $0B;
  fli_LineCompression = $0C;
  fli_Black           = $0D;
  fli_Brun            = $0F;
  fli_UnCompressed    = $10;
  flc_PictureStamp    = $12;


var
  FLC         : TBufStream;
  Output      : TBufStream;
  buffer      : word;
  Header      : FLCFileHeader;
  ChunkHeader : FLCChunkHeader;
  FrameHeader : FLCFrameHeader;
  i           : word;
  Palette     : array [byte,1..3] of byte;
  Pos         : longint;

procedure Error(S : string);
begin
  writeln(S);
  halt;
end;

procedure UnpackPalette(ShiftRight : byte);
var
  Count : word;
  Index : byte;
  i     : byte;
  D     : byte;
  C     : byte;
begin
  if ShiftRight=0 then writeln('  FLI_Palette64')
  else writeln('  FLC_Palette256');
  FLC.Read(Count,2);
  if Count<>0 then
   begin
     Index:=0;
     for i:=1 to Count do
      begin
        FLC.Read(D,1);
        FLC.Read(C,1);
        Inc(Index,D);
        if C=0 then
         begin
           FLC.Read(Palette[Index],256*3);
           Inc(Index,256);
         end
        else
         begin
           FLC.Read(Palette[Index],C*3);
           Inc(Index,C);
         end;
      end;
     for i:=0 to 255 do
      begin
        Palette[i,1]:=Palette[i,1] shr ShiftRight;
        Palette[i,2]:=Palette[i,2] shr ShiftRight;
        Palette[i,3]:=Palette[i,3] shr ShiftRight;
      end;
   end;
end;

procedure ConvertBrun;
var
  i,j     : word;
  Counter : byte;
  Num     : byte;
  Buffer  : array[byte] of byte;
  y       : word;
const
  Signature : array [1..4] of char = 'BRUN';
begin
  writeln('  FLI_Brun');
  Output.Init(paramstr(2),stCreate,65535);
  Output.Write(Signature,SizeOf(Signature));
  Output.Write(Header.Width,2);
  Output.Write(Header.Height,2);
  for y:=1 to Header.Height do
   begin
     FLC.Read(Num,1);
     Output.Write(Num,1);
     if Num>0 then
      begin
        for j:=1 to Num do
         begin
           FLC.Read(Counter,1);
           if Counter>$80 then
            begin
              Output.Write(Counter,1);
              Counter:=256-Counter;
              FLC.Read(Buffer,Counter);
              Output.Write(Buffer,Counter);
            end
           else
            begin
              if Counter>0 then
               begin
                 dec(Counter);
                 Output.Write(Counter,1);
                 FLC.Read(Buffer,1);
                 Output.Write(Buffer,1);
              end;
            end;
         end;
        if FLC.Status<>stOk then Error(' FLC file read error.');
      end;
   end;
  Output.Flush;
  if Output.Status<>stOk then Error('Could not save BRUN file.');
  Output.Done;
end;

begin
  Writeln(Ofs(Header.MTime)-Ofs(Header));
  if paramcount<>3 then
   begin
     writeln('FLC2BRUN converter by theAR/TATE.');
     writeln('Use: FLC2BRUN _input_FLC_|_input_ANIMCEL_ _output_BRN_ _output_PAL_');
     halt;
   end;
  FLC.Init(paramstr(1),stOpenRead,65535);
  if FLC.Status<>stOk then Error('File not found.');
  FLC.Read(Header,SizeOf(Header));
  if FLC.Status<>stOk then Error('Error in FLC.');
  if Header.Signature<>$AF12 then Error('Invalid file type.');
  if Header.Status<>$0003 then writeln('Warning: Invalid FLC status.');
  if Header.Depth<>8 then writeln('Warning: Number of colors is not 256.');
  FLC.Seek(Header.FirstFramePos);
  FLC.Read(FrameHeader,SizeOf(FrameHeader));
  if FLC.Status<>stOk then Error('Error in FLC.');
  if FrameHeader.Signature<>$F1FA then Error('Invalid frame in file.');
  if FrameHeader.Chunks=0 then Error('Out of data in frame.');
  for i:=1 to FrameHeader.Chunks do
   begin
     Pos:=FLC.GetPos;
     FLC.Read(ChunkHeader,SizeOf(ChunkHeader));
     case ChunkHeader.Signature of
       flc_Palette256 : UnpackPalette(2);
       fli_Palette64  : UnpackPalette(0);
       fli_Brun       : ConvertBrun;
       else
        FLC.Seek(FLC.GetPos+ChunkHeader.Size-SizeOf(ChunkHeader));
     end;
   end;
  FLC.Done;
  Output.Init(paramstr(3),stCreate,1024);
  Output.Write(Palette,SizeOf(Palette));
  Output.Flush;
  if Output.Status<>stOk then Error('Could not save PAL file.');
  Output.Done;
end.